from datetime import date, datetime
from typing import List

import brewman.schemas.net_transactions as schemas

from fastapi import APIRouter, Request, Security
from sqlalchemy import not_
from sqlalchemy.orm import Session
from sqlalchemy.sql.expression import desc, func, select

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_period
from ...db.session import SessionFuture
from ...models.account_base import AccountBase
from ...models.journal import Journal
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.NetTransactions)
def show_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["net-transactions"]),
) -> schemas.NetTransactions:
    return schemas.NetTransactions(
        startDate=get_start_date(request.session),
        finishDate=get_finish_date(request.session),
        body=[],
    )


@router.get("/{start}/{finish}", response_model=schemas.NetTransactions)
def show_data(
    start: str,
    finish: str,
    request: Request,
    user: UserToken = Security(get_user, scopes=["net-transactions"]),
) -> schemas.NetTransactions:
    set_period(start, finish, request.session)
    with SessionFuture() as db:
        return schemas.NetTransactions(
            startDate=start,
            finishDate=finish,
            body=build_report(
                datetime.strptime(start, "%d-%b-%Y"),
                datetime.strptime(finish, "%d-%b-%Y"),
                db,
            ),
        )


def build_report(start_date: date, finish_date: date, db: Session) -> List[schemas.NetTransactionsItem]:
    amount_sum = func.sum(Journal.amount * Journal.debit).label("amount")
    query = db.execute(
        select(AccountBase, amount_sum)
        .join(Journal.voucher)
        .join(Journal.account)
        .where(
            Voucher.date >= start_date,
            Voucher.date <= finish_date,
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
        )
        .group_by(AccountBase)
        .order_by(AccountBase.type_id, desc(func.abs(amount_sum)))
    ).all()

    body = []
    for account, amount in query:
        if amount != 0:
            item = schemas.NetTransactionsItem(type=account.type_.name, name=account.name)
            if amount > 0:
                item.debit = amount
            else:
                item.credit = amount
            body.append(item)
    return body
