from datetime import date, datetime
from decimal import Decimal
from typing import List, Optional, Union

from pydantic import validator
from pydantic.main import BaseModel

from . import to_camel


class CashFlowItem(BaseModel):
    name: str
    url: Optional[List[str]]
    amount: Decimal

    class Config:
        anystr_strip_whitespace = True
        alias_generator = to_camel


class CashFlowBody(BaseModel):
    operating: List[CashFlowItem]
    investing: List[CashFlowItem]
    financing: List[CashFlowItem]
    details: List[CashFlowItem]

    class Config:
        anystr_strip_whitespace = True
        alias_generator = to_camel


class CashFlow(BaseModel):
    start_date: date
    finish_date: date
    body: CashFlowBody
    footer: List[CashFlowItem]

    class Config:
        anystr_strip_whitespace = True
        alias_generator = to_camel
        json_encoders = {date: lambda v: v.strftime("%d-%b-%Y")}

    @validator("start_date", pre=True)
    def parse_start_date(cls, value: Union[date, str]) -> date:
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()

    @validator("finish_date", pre=True)
    def parse_finish_date(cls, value: Union[date, str]) -> date:
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()
