import { DataSource } from '@angular/cdk/collections';
import { EventEmitter } from '@angular/core';
import { MatPaginator, PageEvent } from '@angular/material/paginator';
import { MatSort, Sort } from '@angular/material/sort';
import { merge, Observable, of as observableOf } from 'rxjs';
import { map, tap } from 'rxjs/operators';

import { Product } from '../../core/product';
import { ProductGroup } from '../../core/product-group';

export class ProductListDataSource extends DataSource<Product> {
  private filterValue = '';

  constructor(
    public data: Product[],
    private readonly filter: Observable<string>,
    private paginator?: MatPaginator,
    private sort?: MatSort,
  ) {
    super();
    this.filter = filter.pipe(
      tap((x) => {
        this.filterValue = x;
      }),
    );
  }

  connect(): Observable<Product[]> {
    const dataMutations: (
      | Observable<Product[]>
      | Observable<string>
      | EventEmitter<PageEvent>
      | EventEmitter<Sort>
    )[] = [observableOf(this.data), this.filter];
    if (this.paginator) {
      dataMutations.push((this.paginator as MatPaginator).page);
    }
    if (this.sort) {
      dataMutations.push((this.sort as MatSort).sortChange);
    }

    return merge(...dataMutations).pipe(
      map(() => this.getFilteredData([...this.data])),
      tap((x: Product[]) => {
        if (this.paginator) {
          this.paginator.length = x.length;
        }
      }),
      map((x: Product[]) => this.getPagedData(this.getSortedData(x))),
    );
  }

  disconnect() {}

  private getFilteredData(data: Product[]): Product[] {
    return this.filterValue.split(' ').reduce(
      (p: Product[], c: string) =>
        p.filter((x) => {
          const productString = `${x.code} ${x.name} ${x.productGroup?.name}${
            x.isPurchased ? ' purchased' : ' made'
          }${x.isSold ? 'sold' : 'used'}${x.isActive ? 'active' : 'deactive'}`.toLowerCase();
          return productString.indexOf(c) !== -1;
        }),
      Object.assign([], data),
    );
  }

  private getPagedData(data: Product[]) {
    if (this.paginator === undefined) {
      return data;
    }
    const startIndex = this.paginator.pageIndex * this.paginator.pageSize;
    return data.splice(startIndex, this.paginator.pageSize);
  }

  private getSortedData(data: Product[]) {
    if (this.sort === undefined) {
      return data;
    }
    if (!this.sort.active || this.sort.direction === '') {
      return data;
    }

    const sort = this.sort as MatSort;
    return data.sort((a, b) => {
      const isAsc = sort.direction === 'asc';
      switch (sort.active) {
        case 'name':
          return compare(a.name, b.name, isAsc);
        case 'productGroup':
          return compare(
            (a.productGroup as ProductGroup).name,
            (b.productGroup as ProductGroup).name,
            isAsc,
          );
        default:
          return 0;
      }
    });
  }
}

/** Simple sort comparator for example ID/Name columns (for client-side sorting). */
const compare = (a: string | number, b: string | number, isAsc: boolean) =>
  (a < b ? -1 : 1) * (isAsc ? 1 : -1);
